-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------

NUM_SPELL_SLOTS = 10;
VERSION = "2.0.3.5"

-------------------------------------------------------------------------------
-- Variables
-------------------------------------------------------------------------------

SHOW_WELCOME = true;
FLOTOTEMBAR_LAYOUT = "auto";
FLOTOTEMBAR_BUTTONS_ORDER = {};
FLO_CLASS_NAME = nil;

-------------------------------------------------------------------------------
-- Functions
-------------------------------------------------------------------------------

-- Executed on load, calls general set-up functions
function FloTotemBar_OnLoad()

	-- Class-based setup, abort if not supported
	_, FLO_CLASS_NAME = UnitClass("player");
	FLO_CLASS_NAME = strupper(FLO_CLASS_NAME);

	local classSpells = FLO_TOTEM_SPELLS[FLO_CLASS_NAME];

	if ( classSpells == nil ) then
		return;
	end

	this.totemtype = string.sub(this:GetName(), 7);

	-- Store the spell list for later
	this.AvailableSpells = classSpells[this.totemtype];
	if ( this.AvailableSpells == nil ) then
		return;
	end

	-- Init the order variable
	FLOTOTEMBAR_BUTTONS_ORDER[this.totemtype] = {};

	this.Spells = {};

	-- Cosmetic
	this:SetBackdropBorderColor(0.5, 0.5, 0.5);
	this:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b, 0.7);

	if ( SHOW_WELCOME ) then
		DEFAULT_CHAT_FRAME:AddMessage( "FloTotemBar "..VERSION.." loaded." );
		SHOW_WELCOME = nil;

		SLASH_FLOTOTEMBAR1 = "/flototembar";
		SLASH_FLOTOTEMBAR2 = "/ftb";
		SlashCmdList["FLOTOTEMBAR"] = FloTotemBar_SetLayout;

		this:RegisterEvent("VARIABLES_LOADED");
	end
	this:RegisterEvent("PLAYER_ENTERING_WORLD");
	this:RegisterEvent("LEARNED_SPELL_IN_TAB");
	this:RegisterEvent("PLAYER_ALIVE");
	this:RegisterEvent("PLAYER_LEVEL_UP");
	this:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED");
	this:RegisterEvent("SPELL_UPDATE_COOLDOWN");
	this:RegisterEvent("SPELL_UPDATE_USABLE");
	this:RegisterEvent("PLAYER_DEAD");
	this:RegisterEvent("UPDATE_BINDINGS");

	-- Destruction detection
	if ( FLO_CLASS_NAME == "HUNTER" ) then
		this:RegisterEvent("CHAT_MSG_SPELL_PERIODIC_CREATURE_DAMAGE");
		this:RegisterEvent("CHAT_MSG_SPELL_PERIODIC_HOSTILEPLAYER_DAMAGE");
		this:RegisterEvent("CHAT_MSG_SPELL_PET_DAMAGE");
		this:RegisterEvent("CHAT_MSG_SPELL_SELF_DAMAGE");
	else
		this:RegisterEvent("CHAT_MSG_SPELL_CREATURE_VS_SELF_DAMAGE");
		this:RegisterEvent("CHAT_MSG_COMBAT_CREATURE_VS_SELF_HITS");
		this:RegisterEvent("CHAT_MSG_SPELL_HOSTILEPLAYER_DAMAGE");
		this:RegisterEvent("CHAT_MSG_COMBAT_HOSTILEPLAYER_HITS");
	end
end

function FloTotemBar_OnEvent(event, text, spell, rank)

	if ( event == "PLAYER_ENTERING_WORLD" or event == "LEARNED_SPELL_IN_TAB" or event == "PLAYER_ALIVE" or event == "PLAYER_LEVEL_UP" ) then
		FloTotemBar_Setup();
	elseif ( event == "UNIT_SPELLCAST_SUCCEEDED" ) then
		if ( text == "player" ) then
			FloTotemBar_StartTimer(spell, rank);
		end
	elseif ( event == "SPELL_UPDATE_COOLDOWN" or event == "SPELL_UPDATE_USABLE" ) then
		FloTotemBar_UpdateState();
	elseif ( event == "PLAYER_DEAD" ) then
		FloTotemBar_ResetTimer();
	elseif ( event == "VARIABLES_LOADED" ) then

		FloTotemBar_SetLayout(FLOTOTEMBAR_LAYOUT);

		-- Hook the UIParent_ManageFramePositions function
		hooksecurefunc("UIParent_ManageFramePositions", FloTotemBar_UpdatePosition);
	elseif ( event == "UPDATE_BINDINGS" ) then
		FloLib_UpdateBindings(this, "FLOTOTEM"..this.totemtype);
	else
		-- Chat events used for totem destruction detection
		if ( this.activeSpell ) then
			this.Spells[this.activeSpell].algo(event, text);
		end
	end
end

function FloTotemBar_ReceiveDrag(releaseCursor)

	if ( InCombatLockdown() ) then
		return;
	end

	local cursorType, index, info;

	cursorType, index, info = GetCursorInfo();

	if ( cursorType ~= "spell" or info ~= BOOKTYPE_SPELL ) then
		return;
	end

	local button = this;
	local newspell = GetSpellName(index, info);
	this = this:GetParent();

	-- find the spell in the curent list
	for i = 1, #this.AvailableSpells do
		if ( this.AvailableSpells[i].name == newspell ) then
			if ( releaseCursor ) then
				ClearCursor();
			end
			FloLib_Swap(FLOTOTEMBAR_BUTTONS_ORDER[this.totemtype], FLOTOTEMBAR_BUTTONS_ORDER[this.totemtype][button:GetID()], i);
			FloTotemBar_Setup();
			break;
		end
	end

end

function FloTotemBar_CheckTotemLife(event, text)

	local name, damages;

	text = string.upper(text);
	name = string.upper(this.Spells[this.activeSpell].name);

	damages = string.match(text, name.."[^%d]*(%d+)");
	if ( damages ) then

		this.life = this.life - damages;

		if ( this.life <= 0 ) then
			FloTotemBar_ResetTimer();
		end
	end
end

function FloTotemBar_CheckTrapLife(event, text)

	local name;

	text = string.upper(text);
	name = string.upper(this.Spells[this.activeSpell].name);

	if ( event == "CHAT_MSG_SPELL_PET_DAMAGE" ) then
		if ( string.find(text, name, 1, true) ) then
			FloTotemBar_ResetTimer();
		end
	elseif ( event == "CHAT_MSG_SPELL_PERIODIC_CREATURE_DAMAGE" or event == "CHAT_MSG_SPELL_PERIODIC_HOSTILEPLAYER_DAMAGE" ) then
		if ( string.find(text, name, 1, true) ) then
			FloTotemBar_TimerRed();
		end
	end
end

function FloTotemBar_CheckTrap2Life(event, text)

	local name;
	if ( event == "CHAT_MSG_SPELL_SELF_DAMAGE" ) then
		text = string.upper(text);
		name = string.upper(this.Spells[this.activeSpell].name);
		if ( string.find(text, name, 1, true) ) then
			FloTotemBar_ResetTimer();
		end
	end
end

function FloTotemBar_SetupSpell(spell, spellID, pos)

	local durationMod = 0;
	local lifeMod = 0;
	local algo, lifeTable;
	local button, icon, texture;

	button = getglobal(this:GetName().."Button"..pos);
	icon = getglobal(this:GetName().."Button"..pos.."Icon");

	button:SetAttribute("type1", "spell");
	button:SetAttribute("spell", spell.name);

	texture = GetSpellTexture(spellID, BOOKTYPE_SPELL);
	icon:SetTexture(texture);

	if ( spell.firetalent ) then
		durationMod = FloTotemBar_GetTalentRank(FLO_IMPROVED_FIRE_TALENT, 1);
	end

	if ( spell.lifetable ) then
		lifeTable = {};
		lifeMod = FloTotemBar_GetTalentRank(FLO_EARTHS_GRASP_TALENT, 1);
		for rank, value in pairs(spell.lifetable) do
			lifeTable[rank] = (value + 3 * UnitLevel("player")) * (1 + 0.25 * lifeMod);
		end
	end

	if ( FLO_CLASS_NAME ~= "HUNTER" ) then
		algo = FloTotemBar_CheckTotemLife;
	elseif ( spell.x ) then
		algo = FloTotemBar_CheckTrap2Life;
	else
		algo = FloTotemBar_CheckTrapLife;
	end

	this.Spells[pos] = { name = spell.name, id = spellID, duration = spell.duration + durationMod, k = spell.k, algo = algo, lifetable = lifeTable, specialduration = spell.specialduration };

end

function FloTotemBar_Setup()

	local numSpells = 0;
	local button;
	local spellID, spell;
	local i = 1;

	this.Spells = {};

	-- Check already positionned spells
	while ( FLOTOTEMBAR_BUTTONS_ORDER[this.totemtype][i] ) do

		local n = FLOTOTEMBAR_BUTTONS_ORDER[this.totemtype][i];

		spell = this.AvailableSpells[n];
		spellID = FloTotemBar_GetSpellID(spell.name);
		if ( spellID > -1 ) then
			FloTotemBar_SetupSpell(spell, spellID, i);
			i = i + 1;
		else
			-- this spell is unavailable, shift the remaining indexes by 1
			for j = i, #FLOTOTEMBAR_BUTTONS_ORDER[this.totemtype] do
				FLOTOTEMBAR_BUTTONS_ORDER[this.totemtype][j] = FLOTOTEMBAR_BUTTONS_ORDER[this.totemtype][j+1];
			end
		end

	end

	numSpells = i - 1;

	for n = 1, #this.AvailableSpells do

		if ( n > NUM_SPELL_SLOTS ) then
			break;
		end

		-- Check if this spell is already positionned
		i = nil;
		for j = 1, #FLOTOTEMBAR_BUTTONS_ORDER[this.totemtype] do
			if ( FLOTOTEMBAR_BUTTONS_ORDER[this.totemtype][j] == n ) then
				i = 1;
				break;
			end
		end

		if ( not i ) then
			spell = this.AvailableSpells[n];
			spellID = FloTotemBar_GetSpellID(spell.name);
			if ( spellID > -1 ) then

				numSpells = numSpells + 1;

				FloTotemBar_SetupSpell(spell, spellID, numSpells);
				FLOTOTEMBAR_BUTTONS_ORDER[this.totemtype][numSpells] = n;
			end
		end
	end

	if ( numSpells > 0 ) then

		this:Show();
		this:SetWidth(numSpells * 35 + 27);

		for i=1, NUM_SPELL_SLOTS do

			button = getglobal(this:GetName().."Button"..i);

			if ( i <= numSpells ) then
				button:Show();
			else
				button:Hide();
			end
		end
	else
		this:Hide();
	end
	FloTotemBar_UpdateState();
end

function FloTotemBar_UpdateState()

	local numSpells = #this.Spells;
	local spell, cooldown, normalTexture, icon;
	local start, duration, enable, isUsable, noMana;

	for i=1, numSpells do

		spell = this.Spells[i];

		--Cooldown stuffs
		cooldown = getglobal(this:GetName().."Button"..i.."Cooldown");
		start, duration, enable = GetSpellCooldown(spell.id, BOOKTYPE_SPELL);
		CooldownFrame_SetTimer(cooldown, start, duration, enable);

		--Castable stuffs
		normalTexture = getglobal(this:GetName().."Button"..i.."NormalTexture");
		icon = getglobal(this:GetName().."Button"..i.."Icon");
		isUsable, noMana = IsUsableSpell(spell.id, BOOKTYPE_SPELL);

		if ( isUsable ) then
			icon:SetVertexColor(1.0, 1.0, 1.0);
			normalTexture:SetVertexColor(1.0, 1.0, 1.0);
		elseif ( noMana ) then
			icon:SetVertexColor(0.5, 0.5, 1.0);
			normalTexture:SetVertexColor(0.5, 0.5, 1.0);
		else
			icon:SetVertexColor(0.4, 0.4, 0.4);
			normalTexture:SetVertexColor(1.0, 1.0, 1.0);
		end

	end

end

function FloTotemBar_UpdatePosition()

	if ( FLOTOTEMBAR_LAYOUT ~= "auto" ) then
		return;
	end

	local yOffset = -3;
	local yOffset1 = 0;
	local yOffset2 = 0;
	local anchorFrame;

	if ( not MainMenuBar:IsShown() ) then
		anchorFrame = "UIParent";
		yOffset = 110-UIParent:GetHeight();
	else
		anchorFrame = "MainMenuBar";
		if ( ReputationWatchBar:IsShown() and MainMenuExpBar:IsShown() ) then
			yOffset = yOffset + 9;
		end

		if ( MainMenuBarMaxLevelBar:IsShown() ) then
			yOffset = yOffset - 5;
		end

		if ( SHOW_MULTI_ACTIONBAR_2 ) then
			yOffset2 = yOffset2 + 45;
		end

		if ( SHOW_MULTI_ACTIONBAR_1 ) then
			yOffset1 = yOffset1 + 45;
		end
	end

	if ( FLO_CLASS_NAME == "HUNTER" ) then
		if ( FloAspectBar) then
			FloBarEARTH:SetPoint("TOPLEFT", "FloAspectBar", "TOPRIGHT", 12, 0);
		else
			FloBarEARTH:SetPoint("TOPLEFT", anchorFrame, "TOPLEFT", 512, FloBarEARTH:GetHeight() + yOffset + yOffset2);
		end
	else
		FloBarEARTH:SetPoint("TOPLEFT", anchorFrame, "TOPLEFT", 6, FloBarEARTH:GetHeight() + yOffset + yOffset1);
		FloBarFIRE:SetPoint("TOPLEFT", "FloBarEARTH", "TOPRIGHT", 6, 0);
		FloBarWATER:SetPoint("TOPLEFT", "FloBarFIRE", "TOPRIGHT", 6, 0);
		FloBarAIR:SetPoint("TOPLEFT", "FloBarWATER", "TOPRIGHT", 6, 0);
	end
end

function FloTotemBar_SetBarDrag(frame, enable)

	local countdown = getglobal(frame:GetName().."Countdown");
	frame:EnableMouse(enable);
	if ( enable ) then
		frame:RegisterForDrag("LeftButton");
		countdown:RegisterForDrag("LeftButton");
	end
end

function FloTotemBar_SetLayout(mode)

	if ( mode ~= "lock" and mode ~= "unlock" and mode ~= "auto" ) then
		DEFAULT_CHAT_FRAME:AddMessage( "FloTotemBar options : lock|unlock|auto" );
		return;
	end

	local unlocked = (mode == "unlock");

	FLOTOTEMBAR_LAYOUT = mode;
	DEFAULT_CHAT_FRAME:AddMessage( "FloTotemBar layout "..FLOTOTEMBAR_LAYOUT );

	for i, v in ipairs({FloBarEARTH, FloBarFIRE, FloBarWATER, FloBarAIR}) do
		FloTotemBar_SetBarDrag(v, unlocked);
	end

	if ( mode == "auto" ) then
		-- Force the auto positionning
		FloTotemBar_UpdatePosition();
	end
end

-- Return the rank of a talent
function FloTotemBar_GetTalentRank(talentName, tree)

	local nt = GetNumTalents(tree);
	local n, r, m;

	for i = 1, nt do
		n, _, _, _, r, m = GetTalentInfo(tree, i);
		if ( n == talentName ) then
			return r, m;
		end
	end
	return 0, 0;
end

-- Returns the ID of the given spell from the spellbook
function FloTotemBar_GetSpellID(spell)

	local i = 1;
	local valid = -1;
	local spellName, spellRank = GetSpellName(i, BOOKTYPE_SPELL)
	while spellName do
		if ( spellName == spell) then
			valid = i;
		elseif ( valid > -1 ) then
			return valid;
		end
		i = i + 1;
		spellName, spellRank = GetSpellName(i, BOOKTYPE_SPELL)
	end
	return valid;
end

function FloTotemBar_ResetTimer()

	this.startTime = 0;
	FloTotemBar_OnUpdate();
end

function FloTotemBar_TimerRed()

	local countdown = getglobal(this:GetName().."Countdown");
	countdown:SetStatusBarColor(1.0, 0.0, 0.0);

end

function FloTotemBar_StartTimer(spellName, rank)

	local founded = false;
	local bar;
	local duration;
	local countdown;
	local life = 5;

	-- Special case for Totemic Call
	if ( spellName == FLO_TOTEMIC_CALL_SPELL ) then
		FloTotemBar_ResetTimer();
		return;
	end

	-- Find spell
	for i = 1, #this.Spells do
		if ( this.Spells[i].name == spellName ) then
			founded = i;

			if ( rank == "" and this.Spells[i].specialduration ) then
				duration = this.Spells[i].specialduration;
			else
				duration = this.Spells[i].duration;
			end

			if ( this.Spells[i].k ) then
				duration = duration + this.Spells[i].k * string.sub(rank, -2);
			end

			if ( this.Spells[i].lifetable ) then
				life = this.Spells[i].lifetable[rank];
			end
			break;
		end
	end

	if ( founded ) then

		this.activeSpell = founded;
		this.startTime = GetTime();
		this.life = life;

		countdown = getglobal(this:GetName().."Countdown");
		countdown:SetMinMaxValues(0, duration);
		countdown:SetStatusBarColor(1.0, 0.7, 0.0);
		FloTotemBar_OnUpdate();

	end

end

function FloTotemBar_OnUpdate()

	local isActive;
	local button;
	local countdown;
	local timeleft;
	local duration;

	for i=1, #this.Spells do

		button = getglobal(this:GetName().."Button"..i);

		spell = this.Spells[i];

		if ( this.activeSpell == i ) then

			countdown = getglobal(this:GetName().."Countdown");
			_, duration = countdown:GetMinMaxValues();

			timeleft = this.startTime + duration - GetTime();
			isActive = timeleft > 0;

			if (isActive) then
				countdown:SetValue(timeleft);
			else
				this.activeSpell = nil;
				countdown:SetValue(0);
			end;
		else
			isActive = false;
		end

		if ( isActive ) then
			button:SetChecked(1);
		else
			button:SetChecked(0);
		end
	end
end

